all_inlays = {
    'TIKI_117': {'inlay': 'TIKI_117', 'number': '117', 'number_of_rows': 5, 'received_channel': '37',
                 'energy_pattern_val': 51, 'time_profile_val': [5, 40], 'symbol_val': '1Mhz'},
    'TIKI_121': {'inlay': 'TIKI_121', 'number': '121', 'number_of_rows': 6, 'received_channel': '37',
                 'energy_pattern_val': 51, 'time_profile_val': [5, 40], 'symbol_val': '1Mhz'},
    'TIKI_169': {'inlay': 'TIKI_169', 'number': '169', 'number_of_rows': 4, 'received_channel': '4',
                 'energy_pattern_val': 18, 'time_profile_val': [5, 40], 'symbol_val': '1Mhz'},
    'TIKI_169_68': {'inlay': 'TIKI_169', 'number': '169', 'number_of_rows': 4, 'received_channel': '0',
                    'energy_pattern_val': 18, 'time_profile_val': [5, 40], 'symbol_val': '1Mhz'},
    'TIKI_169_BLE5_C': {'inlay': 'TIKI_169', 'number': '169', 'number_of_rows': 4, 'received_channel': '10',
                        'energy_pattern_val': 18, 'time_profile_val': [5, 40], 'symbol_val': '2Mhz'},
}
